@@ -1,5 +1,5 @@
 char *Version = 
-"tircproxy v0.4.3pre2, Copyleft 1998, Bjarni R. Einarsson <bre@mmedia.is>" ;
+"tircproxy v0.4.5, Copyleft 2000, Bjarni R. Einarsson <bre@netverjar.is>" ;
 
 /*****************************************************************************\
  * 
@@ -14,7 +14,13 @@ char *Version =
  * none at all.  This code is distributed under the GPL.
  *
  * See the files README, CHANGELOG.txt and BUGS.txt for more info!
- * Don't forget to edit config.h to suit your tastes!
+ * Don't forget to edit tircproxy.h to suit your tastes!
+ * 
+ * ************************************************************************* *
+ * 
+ * Note to hackers, porters, etc.
+ * 
+ *   Please, please let me know if you find bugs in my code!
  * 
 \*****************************************************************************/
 
@@ -85,6 +91,18 @@ char *Version =
 # include <crypt.h>
 #endif
 
+#if (HAVE_UDB_H && USE_UDB)
+# warning Using UDB style IPC for identd etc.
+# include <udb.h>
+# undef CDIR
+# define CDIR 1
+# undef CDIR_IDENT
+# define CDIR_IDENT 1
+# undef IP_TO_UID_PREFIX
+#else
+# undef USE_UDB
+#endif
+
 #ifndef INADDR_NONE
 # define INADDR_NONE -1
 #endif
@@ -107,7 +125,11 @@ char *Version =
 #  include <net/if.h>
 #  include <netinet/in_systm.h>
 #  include <netinet/tcp.h>
-#  include <netinet/ip_compat.h>
+#  if HAVE_NETINET_IP_FIL_COMPAT_H
+#	include <netinet/ip_fil_compat.h>
+#  else	
+#	include <netinet/ip_compat.h>
+#  endif
 #  include <netinet/ip_fil.h>
 #  include <netinet/ip_proxy.h>
 #  include <netinet/ip_nat.h>
@@ -130,7 +152,7 @@ char *Version =
 #  define LINUX 0
 #  warning Linux support not available
 # else
-#  warning Linux transparent proxying available
+#  warning Linux transparent proxying available (depends on kernel)
 #  define TRANS 1
 # endif
 #endif
@@ -141,6 +163,16 @@ char *Version =
 # warning Quiz mode not available
 #endif
 
+/* Some useful definitions, which we might get from the system.
+*/
+#ifndef MAXLOGNAME
+# define MAXLOGNAME 64
+#endif
+
+#ifndef PATH_MAX
+# define PATH_MAX 256
+#endif
+
 /* Do we want to use the TCP wrappers for access control?
 */
 #ifdef TCP_WRAPPERS
@@ -164,7 +196,7 @@ typedef unsigned long	ipaddr_t;
 /* Macros & definitions.
 */
 #define FD_MAX(a,b)		((a) > (b) ? (a) : (b))
-#define MIN			60
+#define MINUTE			60
 
 #define DEBUG_FEATURES		1
 #define DEBUG_BASIC		2
@@ -175,10 +207,10 @@ typedef unsigned long	ipaddr_t;
 /* Function prototypes.
 */
 static void usage		(char *prog, char *opt);
-static short get_port_num		(char *portnum);
+static short get_port_num	(char *portnum);
 static ipaddr_t get_ip_addr	(char *ipaddr);
-static uid_t get_user_id		(char *user);
-static uid_t get_group_id		(uid_t uid);
+static uid_t get_user_id	(char *user);
+static uid_t get_group_id	(uid_t uid);
 static int bind_to_port		(ipaddr_t bind_ip, short bind_port, 
 				 int backlog, int dlev);
 static int connect_to_server	(struct sockaddr_in addr);
@@ -202,8 +234,9 @@ static int scan_line		(char *line, int isclient);
 static int dcc_mangle_filename	(char *filename);
 static int dcc_resume		(int destport, int resume); 
 static char *proxy_dcc		(int destaddr, int destport, int incoming);
-static void change_uid		(struct sockaddr_in *addr);
-#ifdef CDIR_IDENT
+static void get_user_name	(struct sockaddr_in *addr);
+static void change_uid		(void);
+#ifdef CDIR
 static void tell_identd		(struct sockaddr_in *, struct sockaddr_in *);
 static void cleanup_identd	(void);
 #endif
@@ -227,12 +260,13 @@ static void quiz_greet		(void);
 
 /* These make useful global variables..
 */
-char		user_nick[LENGTH_NICKNAME * 2], user_name[512], alarm_in[64];
-char		user_ident_file[64];
-char		server_tag[50];
+char		user_nick[LENGTH_NICKNAME * 2];
+char            user_name[MAXLOGNAME];
+char		user_ident_file[PATH_MAX];
+char		server_tag[64], alarm_in[64];
 ipaddr_t	clients_ip = INADDR_NONE;
 char		clients_ip_s[64];
-
+int		motd_is_done = 0;
 
 /* We need to remember some stuff to support DCC RESUME.
 ** See: http://www.mirc.co.uk/help/dccresum.txt 
@@ -256,7 +290,7 @@ char			*from_cli_first = NULL;
 char			**from_cli_last = NULL;
 char			*to_cli_first = NULL;
 char			**to_cli_last = NULL;
-char			quizfile[64], quiz[512];
+char			quizfile[PATH_MAX], quiz[512];
 int			use_unix_passwd = QUIZ_OFF;
 int			use_quiz_mode = QUIZ_OFF;
 #endif
@@ -296,6 +330,9 @@ int		use_tcp_wrappers = 1;
 #ifdef CDIR
 int	   	use_cdir = 1;
 #endif
+#ifdef USE_UDB
+struct udb_connection conn;
+#endif
 
 
 /* Main!
@@ -313,7 +350,7 @@ int main(int argc, char **argv)
 
 	/* Parse the command line arguments.
 	*/
-	while ((arg = getopt(argc, argv, "ab:d:h?i:o:pq:r:s:t:CDHKLMNOQRSU")) != EOF)
+	while ((arg = getopt(argc, argv, "ab:d:h?i:o:pq:r:s:t:CDHIKLMNOQRSU")) != EOF)
 	{
 	   	switch (arg)
 		{
@@ -325,7 +362,7 @@ int main(int argc, char **argv)
                     	bind_ip = get_ip_addr(optarg);
 			break;
 		case 'd':
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 		   	debug_level = atoi(optarg);
 #else
 		   	usage(argv[0],"Feature DEBUG not available.");
@@ -353,7 +390,8 @@ int main(int argc, char **argv)
 		case 'q':
 #ifdef QUIZ_MODE
 		   	use_quiz_mode = QUIZ_ON;
-		   	sprintf(quizfile, "%.63s", optarg);
+		   	strncpy(quizfile, optarg, PATH_MAX);
+			quizfile[PATH_MAX-1] = '\0';
 #else
 		   	usage(argv[0],"Feature QUIZ_MODE not active.");
 #endif
@@ -387,6 +425,14 @@ int main(int argc, char **argv)
 			use_tcp_wrappers = 0;
 #else
 		   	usage(argv[0],"Feature TCP_WRAPPERS not active.");
+#endif
+		   	break;
+		case 'I':
+		case 'O': /* backwards compatibility ... */
+#ifdef CDIR
+		   	use_cdir = 0;
+#else
+		   	usage(argv[0],"Neither CDIR nor USE_UDB compiled in, sorry.");
 #endif
 		   	break;
 		case 'K':
@@ -406,13 +452,6 @@ int main(int argc, char **argv)
 		case 'N':
 		   	not_an_irc_proxy = 1;
 		   	break;
-		case 'O':
-#ifdef CDIR
-		   	use_cdir = 0;
-#else
-		   	usage(argv[0],"Feature CDIR (oidentd) not active.");
-#endif
-		   	break;
 		case 'R':
 #ifdef PARANOID
 		   	use_paranoid = 0;
@@ -462,32 +501,41 @@ int main(int argc, char **argv)
 	   	if ((visible_ip_i == INADDR_ANY) &&
 		    (allow_dcc_send | allow_dcc_chat | allow_dcc_unknown) &&
 		    (!not_an_irc_proxy))
-	     	{
+		{
 		   	usage(argv[0], "No internal address specified, DCC would break!");
 		}
 
-	   	fprintf(stderr,"No remote server specified, transparent operation assumed.\n");
+	   	fprintf(stderr,
+				"No remote server specified, transparent operation assumed.\n");
 #endif
 	}
    	else
-     	{
+	{
 		if (server_port == -1)
 		{
 		   	server_port = 6667;
 		   	fprintf(stderr,"No remote port specified, defaulting to 6667\n");
 		}
-     	}
+	}
    
    	/* Create a "server tag" for this proxy.  This is used for preventing
 	** loops, while allowing proxy-to-proxy operation.
 	*/
-     	{
+	{
 	   	char hostname[128];
 	   	gethostname(hostname, 127);
-   		sprintf(server_tag,"X-tircproxy[%d/%.10s]\n", 
-			getpid(), hostname );
+
+		/* server_tag is 64 bytes - this is safe */
+   		sprintf(server_tag, "X-tircproxy[%d/%.10s]\n", 
+				getpid(), hostname );
 	}
-	   
+	
+#ifdef USE_UDB	
+	/* Aquire handles to the UDB shared memory.
+	*/
+	udb_init(UDB_ENV_BASE_KEY);
+#endif
+	
 	/* See if we should run as a server.
 	*/
 	if (run_as_server)
@@ -511,7 +559,7 @@ int main(int argc, char **argv)
 		/* Start a server process. When DEBUG is defined, and the
 		** debug level is high, just run in the foreground.
 		*/
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 		switch ((debug_level >= DEBUG_NOFORK) ? 0 : fork())
 #else
 		switch (fork())
@@ -531,7 +579,7 @@ int main(int argc, char **argv)
 			/* Ignore some signals.
 			*/
 			signal(SIGHUP, SIG_IGN);
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 		   	if (debug_level < DEBUG_NOFORK)
 #else
 				signal(SIGINT, SIG_IGN);
@@ -622,7 +670,7 @@ static void usage(char *prog, char *opt)
 Usage: %s [switches..] [options..] [ircserver [ircport]]\n\
 \n\
 Options:\n",Version,prog);
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 	fprintf(stderr,
 "   -d level    Set debug level (0=nothing .. 8=everything, 9=foreground).\n");
 #endif
@@ -643,13 +691,13 @@ Options:\n",Version,prog);
 Switches:\n\
    -M          Disable DCC SEND filename mangling/censorship.\n\
    -C          Disallow DCC CHAT.\n\
-   -S          Disallow DCC file transmissinos (SEND, TSEND, etc.).\n\
+   -S          Disallow DCC file transmissions (SEND, TSEND, etc.).\n\
    -U          Disallow unknown DCC requests.\n\
    -L          Log to stderr instead of syslog.\n\
    -N          Not an IRC proxy, disable all neat-o IRC stuff\n");
 #ifdef CDIR
 	fprintf(stderr,"\
-   -O          Do not attempt to communicate with oidentd.\n");
+   -I          Do not attempt to communicate with the identd.\n");
 #endif
 #ifdef PARANOID
 	fprintf(stderr,"\
@@ -713,7 +761,7 @@ static short get_port_num(char *portnum)
 		endservent();
 	}
 
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
    	debug_msg(DEBUG_TRIVIA, LOG_DEBUG,
 		  "Port lookup %s -> %hd\n", portnum, port);
 #endif
@@ -739,7 +787,7 @@ static ipaddr_t get_ip_addr(char *ipaddr)
 		endhostent();
 	}
 
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
    	debug_msg(DEBUG_BASIC, LOG_DEBUG, 
 		  "IP lookup %s -> 0x%08lx\n", ipaddr, ip);
 #endif
@@ -768,7 +816,7 @@ static uid_t get_user_id(char *user)
 		uid = -1;
 	}
 
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
    	debug_msg(DEBUG_BASIC, LOG_DEBUG,
 	    "User lookup %s -> %d\n", user, uid);
 #endif
@@ -795,7 +843,7 @@ static uid_t get_group_id(uid_t uid)
 		gid = -1;
 	}
 
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
    	debug_msg(DEBUG_TRIVIA, LOG_DEBUG, 
 		  "Group lookup %d -> %d\n", uid, gid);
 #endif
@@ -822,7 +870,7 @@ static int bind_to_port(ipaddr_t bind_ip, short bind_port, int backlog, int dlev
 	   	return(-1);
 	}
 
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 	/* Set the SO_REUSEADDR option for debugging.
 	*/
 	if (debug_level >= DEBUG_NOFORK) {
@@ -863,18 +911,18 @@ static int bind_to_port(ipaddr_t bind_ip, short bind_port, int backlog, int dlev
 */
 static int connect_to_server(struct sockaddr_in addr)
 {
-   	struct sockaddr_in			my_addr;
-	int					sock;
+   	struct sockaddr_in my_addr;
+	int	sock;
 
 	/* Allocate a socket.
 	*/
 	if ((sock = socket(AF_INET, SOCK_STREAM, 
-			getprotobyname("tcp")->p_proto)) < 0)
+					   getprotobyname("tcp")->p_proto)) < 0)
 	{
 		debug_msg(0, LOG_WARNING, "socket(): .%256s", strerror(errno));
 		return (-1);
 	}
-   
+	
 	/* Set the address to connect to, and the address to connect from.
 	*/
 	addr.sin_family = AF_INET;
@@ -883,13 +931,12 @@ static int connect_to_server(struct sockaddr_in addr)
 	my_addr.sin_family = AF_INET;
 	my_addr.sin_port = 0;
 	my_addr.sin_addr.s_addr = visible_ip_o;
-
+	
    	/* Bind to the address we want to use.
 	*/
    	if (bind(sock, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0)
-     	{
-		debug_msg(0, LOG_WARNING, "bind(): %d - %.256s", errno, strerror(errno));
-	}
+	  debug_msg(0, LOG_WARNING, "bind(): %d - %.256s", 
+				errno, strerror(errno));
 
    	/* Connect our socket to the above address.
 	*/
@@ -1027,12 +1074,12 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr)
    	/* Give this thing 10 minutes to get started (paranoia).
 	*/
 	signal(SIGALRM, alarm_signal);
-   	alarm(10*MIN);
+   	alarm(10*MINUTE);
    	strcpy(alarm_in, "trans_proxy");
 
-   	/*  Check who the client is, and drop root privs if we have them.
+   	/*  Check who the client is.
 	*/
-   	change_uid(from_addr);
+   	get_user_name(from_addr);
 
 #ifdef TCP_WRAPPERS
 	if (use_tcp_wrappers)
@@ -1124,7 +1171,7 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr)
 	}
    	else
      	{
-	   	/* Ok, not transparent.. lets connect to the specified server.
+	   	/* Ok, not transparent, we'll connect to the specified server.
 		*/
 	   	memset(&to_addr, '\0', sizeof(to_addr));
 		to_addr.sin_family = AF_INET;
@@ -1132,6 +1179,10 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr)
 		to_addr.sin_addr.s_addr = server_ip;
 	}
 
+   	/*  Drop root privs if we have them.
+	*/
+   	change_uid();
+
         signal(SIGHUP, hup_signal);
 
    	debug_msg(0, LOG_DEBUG,
@@ -1171,29 +1222,29 @@ static int copy_loop(int sock, struct sockaddr_in *from_addr,
 	/* Allow 3 minutes for things to get properly started.
 	*/
 	signal(SIGALRM, alarm_signal);
-   	alarm(3*MIN);
+   	alarm(3*MINUTE);
 	strcpy(alarm_in, "copy_loop: starting");
 
    	/* Lookup an ASCII representation of the host's IP address.
-       	*/
-       	lookup_hostname(from_addr, from_host, sizeof(from_host), 0);
-       	lookup_hostname(to_addr, to_host, sizeof(to_host), 0);
+    */
+	lookup_hostname(from_addr, from_host, sizeof(from_host), 0);
+	lookup_hostname(to_addr, to_host, sizeof(to_host), 0);
 
 	/* Read the first couple of lines from the client, take a 
 	** peek.  Check if the server is looping.
 	*/
    	if (is_server_connection)
-     	{
-#ifdef DEBUG
+	{
+#ifdef TIRC_DEBUG
 		debug_msg(DEBUG_BASIC, LOG_DEBUG, 
-			  "Checking if this is a loopy server..");
+				  "Checking if this is a loopy server..");
 #endif
 	   	server = read_len = -1;
 		copy_once(sock, headers, sizeof(headers), "client",
-			  server, ser_left, &ser_ln, "server",
-			  is_server_connection, 1,
-			  &read_len);
-
+				  server, ser_left, &ser_ln, "server",
+				  is_server_connection, 1,
+				  &read_len);
+			
 		if (read_len <= 0) return 0;
 	   
 	   	strncpy(blurb,headers,63);
@@ -1210,46 +1261,47 @@ static int copy_loop(int sock, struct sockaddr_in *from_addr,
    	junk = sizeof(ourend);
 	strcpy(alarm_in, "copy_loop: getsockname 1");
    	getsockname(server, (struct sockaddr *) &ourend, &junk);
+
+#ifdef CDIR_IDENT
+	/* Tell the ident server which user this connection really belongs to.
+	** 
+	** Note: This may be a race between us and the remote IRC server, 
+	**       depending on the implementation of the ident server.
+	*/
+   	if (use_cdir && is_server_connection) tell_identd(&ourend, to_addr);
+#endif
+
 	strcpy(alarm_in, "copy_loop: getsockname 2");
    	getsockname(sock, (struct sockaddr *) &ourend_i, &junk);
 
    	/* Record what IP addresses we're using.
 	*/
 	visible_ip_o = ourend.sin_addr.s_addr;
-       	lookup_hostname(&ourend, visible_ip_o_s, sizeof(visible_ip_o_s), 0);
-
+	lookup_hostname(&ourend, visible_ip_o_s, sizeof(visible_ip_o_s), 0);
+	
    	if (visible_ip_i == INADDR_ANY) 
-     	{
+	{
 		visible_ip_i = ourend_i.sin_addr.s_addr;
-	       	lookup_hostname(&ourend_i, visible_ip_i_s, sizeof(visible_ip_i_s), 0);
+		lookup_hostname(&ourend_i, visible_ip_i_s, sizeof(visible_ip_i_s), 0);
 	}
-
-#ifdef CDIR_IDENT
-	/* Tell the ident server which user this connection really belongs to.
-	** 
-	** Note: this is a race between us and the remote IRC server.  We
-	**       *should* always win it though.
+	
+	/* Log the facts about the connection.
 	*/
-   	if (use_cdir && is_server_connection) tell_identd(&ourend, to_addr);
-#endif
-
-       	/* Log the facts about the connection.
-       	*/
 	debug_msg(0, LOG_INFO, "%.64s -%.64s:%d-> %.64s:%d",
-		from_host, user_name, ntohs(ourend.sin_port), 
-	        to_host, ntohs(to_addr->sin_port));
-
+			  from_host, user_name, ntohs(ourend.sin_port), 
+			  to_host, ntohs(to_addr->sin_port));
+	
 	/* Pass the data we read just now on to the server..
 	*/
    	if (is_server_connection)
-     	{
+	{
 		junk = strlen(server_tag);
 		copy_once(sock, server_tag, 0, "tircproxy",
-			  server, ser_left, &ser_ln, "server",
-			  is_server_connection, 1, &junk);
+				  server, ser_left, &ser_ln, "server",
+				  is_server_connection, 1, &junk);
 		copy_once(sock, headers, sizeof(headers), "client",
-			  server, ser_left, &ser_ln, "server",
-			  is_server_connection, 1, &read_len);
+				  server, ser_left, &ser_ln, "server",
+				  is_server_connection, 1, &read_len);
 		if (read_len <= 0) return 0;
 	}
    
@@ -1268,9 +1320,9 @@ static int copy_loop(int sock, struct sockaddr_in *from_addr,
 		/* Allow 30 minutes for the loop to complete.
 		*/
 		signal(SIGALRM, alarm_signal);
-		alarm(30*MIN);
+		alarm(30*MINUTE);
 	   	strcpy(alarm_in, "copy_loop: waiting");
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 	   	debug_msg(DEBUG_TRIVIA, LOG_DEBUG,"copy_loop: waiting");
 #endif
 		/* Wait for some data to be read.
@@ -1280,7 +1332,7 @@ static int copy_loop(int sock, struct sockaddr_in *from_addr,
 			if (select(max_fd + 1, &read_fd, NULL, NULL, NULL) < 0)
 			{
 				if (errno != EINTR) {
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 					debug_msg(DEBUG_BASIC, LOG_ERR, 
 						  "select(): %.256s", 
 						  strerror(errno));
@@ -1294,7 +1346,7 @@ static int copy_loop(int sock, struct sockaddr_in *from_addr,
 	   	/* We aren't waiting anymore..
 		*/
 	   	strcpy(alarm_in, "copy_loop: copying");
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 	   	debug_msg(DEBUG_TRIVIA, LOG_DEBUG,"copy_loop: copying");
 #endif
 
@@ -1309,7 +1361,7 @@ static int copy_loop(int sock, struct sockaddr_in *from_addr,
 			/* Send the user the MOTD.
 			*/
 			if ((!motd_sent) &&
-			    (copied_bytes > 512) && 
+			    (motd_is_done) && 
 			    (is_server_connection))
 		     	{
 			   	broadcast(sock, IRC_MOTD_FILE);
@@ -1347,7 +1399,7 @@ static int copy_loop(int sock, struct sockaddr_in *from_addr,
 #ifdef QUIZ_MODE
 	   	/* Dump queued data meant for the client.
 		*/
-	   	if (use_quiz_mode)
+	   	if (use_quiz_mode && motd_is_done)
 			quiz_dump_lines(sock, QUIZ_C);
 
 	   	/* Dump queued data, stop quizzing if quiz mode is "flush".
@@ -1379,53 +1431,62 @@ static int  copy_once(int rsock, char *rbuffer, int rbufsize, char *rname,
 { 
    	int rlen;
    	
+#ifdef TIRC_DEBUG
+	debug_msg(DEBUG_TRIVIA, LOG_DEBUG, 
+			  "Entering copy_once (reading %s, writing %s)",
+			  rname, wname);
+#endif
+
    	rlen = *read_len;
    	if (rlen <= 0) rlen = read(rsock, rbuffer, rbufsize - 1);
-   
+
 	switch (rlen)
 	{
-	case -1:
-#ifdef DEBUG
+	  case -1:
+#ifdef TIRC_DEBUG
 		debug_msg(DEBUG_BASIC, 
-			(errno == ECONNRESET ? LOG_DEBUG : LOG_WARNING),
-			"read(%.32s) failed: %.256s", rname, strerror(errno));	
+				  (errno == ECONNRESET ? LOG_DEBUG : LOG_WARNING),
+				  "read(%.32s) failed: %.256s", rname, strerror(errno));	
 #endif
 		if (errno != EINTR)
-	     	{
+		{
 		   	close(wsock);
 			return -2;
 		}
 	   	return 0;
-	case 0:
+	  case 0:
+#ifdef TIRC_DEBUG
+		debug_msg(DEBUG_TRIVIA, LOG_DEBUG, "EOF!");
+#endif
 		close(rsock);
 		close(wsock);
 		return -1;
-	default:
-            	if (*read_len >= 0)
-	     	{
-       	   		if (is_server_connection)
-		     	{
-       			   	if (!filtered_write(wsock, rbuffer, rlen,
-				       	wlbuffer, wlbufpos, r_is_client)) 
-			     	{
-#ifdef DEBUG
+	  default:
+		if (*read_len >= 0)
+		{
+			if (is_server_connection)
+			{
+				if (!filtered_write(wsock, rbuffer, rlen,
+									wlbuffer, wlbufpos, r_is_client)) 
+				{
+#ifdef TIRC_DEBUG
 					debug_msg(DEBUG_BASIC, LOG_WARNING, 
-						"write(%.32s) failed: %.256s",
-						wname, strerror(errno));
+							  "write(%.32s) failed: %.256s",
+							  wname, strerror(errno));
 #endif
 				   	close(rsock);
 				   	return 0;
 				}
 			}
 	   		else
-	     		{
+			{
 				while (write(wsock, rbuffer, rlen) < 0)
-					if (errno != EINTR)
+				  if (errno != EINTR)
 				{
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 					debug_msg(DEBUG_BASIC, LOG_WARNING, 
-						"write(%.32s) failed: %.256s",
-						wname, strerror(errno));
+							  "write(%.32s) failed: %.256s",
+							  wname, strerror(errno));
 #endif
 					close(rsock);
 					return 0;
@@ -1434,6 +1495,11 @@ static int  copy_once(int rsock, char *rbuffer, int rbufsize, char *rname,
 			break;
 		}
 	}
+#ifdef TIRC_DEBUG
+	debug_msg(DEBUG_TRIVIA, LOG_DEBUG, 
+			  "Leaving copy_once (rlen = %d)", rlen);
+#endif
+
    	return (*read_len = rlen);
 }
 
@@ -1448,6 +1514,10 @@ static int filtered_write(int to_sock, char *buff, int blen,
    	char	*nl, *ol, *lp;
   	int	l,silent;
 
+#ifdef TIRC_DEBUG
+	debug_msg(DEBUG_TRIVIA, LOG_DEBUG, "Entering filtered_write");
+#endif
+	
    	buff[blen] = '\0';	/* Null terminate buffer for strchr */
 
 	/* Break the input into lines, parse each one and send them one at
@@ -1456,16 +1526,16 @@ static int filtered_write(int to_sock, char *buff, int blen,
    	nl = ol = buff;
    	lp = leftover + *llen;
    	while ((nl = strchr(ol,'\n')) != NULL)
-     	{
+	{
 	   	/* nl points to the first '\n' in the input.
-	  	** ol points to the beginning of the unscanned input.
+		** ol points to the beginning of the unscanned input.
 	   	** lp points to the end of the contents of 'leftover'
 		**    which is the the fragment of a line we couldn't finish
 		**    processing last time.
 		*/
 
 	   	nl++;			/* Move nl past the '\n'		*/
-	   	l = nl-ol;		/* l = the length of this line		*/
+		l = nl-ol;		/* l = the length of this line		*/
 	   	if (l > 1500) l = 1500;	/* Arbitrary limits == bad!  FIXME!!		*/
 	   	memmove(lp, ol, l);	/* Append new line to leftover		*/
 	   	lp[l] = '\0';		/* Null terminate leftover		*/
@@ -1474,31 +1544,32 @@ static int filtered_write(int to_sock, char *buff, int blen,
 		** buffer - scan it for interesting bits, and send it if it 
 		** isn't deemed completely obscene. :)
 		*/
-	   	if ((silent = scan_line(leftover, isclient)) > 0) {
+	   	if ((silent = scan_line(leftover, isclient)) > 0)
+		{
 			while (write(to_sock, leftover, 
-				     strlen(leftover)) < 0)
-			    	if (errno != EINTR)
+						 strlen(leftover)) < 0)
+			  if (errno != EINTR)
 			{
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 				debug_msg(DEBUG_BASIC, LOG_WARNING, 
-				       "write(%s) failed: %.256s",
-			       	       isclient ? "client" : "server",
-				       strerror(errno));
+						  "write(%s) failed: %.256s",
+						  isclient ? "client" : "server",
+						  strerror(errno));
 #endif
 				close(to_sock);
 				return 0;
 			}
 		}
-#ifdef DEBUG
-	   	else	if (silent >= 0)
-	     			debug_msg(DEBUG_FEATURES, LOG_DEBUG,
-					"Squelched: %.256s", leftover);
+#ifdef TIRC_DEBUG
+			else if (silent >= 0)
+			  debug_msg(DEBUG_FEATURES, LOG_DEBUG,
+						"Squelched: %.256s", leftover);
 #endif
 
    		/* Maintain loop integrity..
 		*/
-	   	ol = nl;
-	   	lp = leftover;
+		ol = nl;
+		lp = leftover;
 	}
    	/* Loop has terminated - if anything is left in ol, then it isn't
 	** a complete line we can scan, it's only a fragment.
@@ -1507,8 +1578,8 @@ static int filtered_write(int to_sock, char *buff, int blen,
 	*/
 	*llen = blen + (buff - ol);
 	memmove(leftover, ol, *llen);
-
-      	return 1;
+	
+	return 1;
 }
 
 
@@ -1563,7 +1634,7 @@ static int scan_line(char *line, int isclient)
 			   	arg_1, &arg_2, &arg_3,
 			   	&scanned) >= 5)
 		     	{
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 			   	debug_msg(DEBUG_FEATURES, LOG_DEBUG,
 				       	"Caught: %.256s", ctcp_begin);
 #endif
@@ -1573,7 +1644,7 @@ static int scan_line(char *line, int isclient)
 			     	{
        					if (!allow_dcc_chat)
 #ifdef DISALLOW_DCC_CHAT
-				     		sprintf(replace, 
+					        sprintf(replace, 
 							DISALLOW_DCC_CHAT, 
 							type, arg_1);
 #else
@@ -1587,7 +1658,7 @@ static int scan_line(char *line, int isclient)
 			     	{
 			     		if (!allow_dcc_send)
 #ifdef DISALLOW_DCC_SEND
-			       			sprintf(replace,
+					        sprintf(replace, 
 							DISALLOW_DCC_SEND,
 							type, arg_1);
 #else
@@ -1595,7 +1666,7 @@ static int scan_line(char *line, int isclient)
 #endif
 				   	else if (!dcc_mangle_filename(arg_1))
 #ifdef MANGLE_DCC_SEND
-			       			sprintf(replace,
+					        sprintf(replace, 
 							MANGLE_DCC_SEND,
 							type, arg_1);
 #else
@@ -1604,7 +1675,7 @@ static int scan_line(char *line, int isclient)
 			     	}
 			   	else if (!strcasecmp(type, "RESUME"))
 			     	{
-				   	sprintf(replace,
+				        sprintf(replace, 
 						"%cDCC %s %s %d %d%s%c",
 						CTRL_A, 
 						type, arg_1,
@@ -1613,7 +1684,7 @@ static int scan_line(char *line, int isclient)
        				}
        			   	else if (!strcasecmp(type, "ACCEPT"))
 			     	{
-				   	sprintf(replace, 
+				        sprintf(replace, 
 						"%cDCC %s %s %d %d%s%c",
        						CTRL_A, 
 						type, arg_1,
@@ -1624,7 +1695,7 @@ static int scan_line(char *line, int isclient)
 				*/
 				else if (!allow_dcc_unknown)
 #ifdef DISALLOW_DCC_SEND
-			       	   	sprintf(replace, 
+				        sprintf(replace, 
 						DISALLOW_DCC_FUNK, 
 						type, arg_1);
 #else
@@ -1636,7 +1707,7 @@ static int scan_line(char *line, int isclient)
 				   	/* No replacement has been set, so
 					** proxy this connection.  :-)
 					*/
-       				   	sprintf(replace,
+				        sprintf(replace, 
 						"%c%s %s %s %s%s%c",
        						CTRL_A, ctcp_type,
 						type, arg_1,
@@ -1652,7 +1723,7 @@ static int scan_line(char *line, int isclient)
 		   	ctcp_begin--;
 		   	if (*replace)
 		     	{
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 			   	debug_msg(DEBUG_FEATURES, LOG_DEBUG,
 					  "Sent: %.256s", replace);
 #endif
@@ -1707,7 +1778,8 @@ static int scan_line(char *line, int isclient)
 		{
 			if ((isdigit(*type) ||
 			    (!strcasecmp(type, "PRIVMSG"))) && 
-			    ((*arg != '&') && (*arg != '#') && 
+			    ((*arg != '&') && (*arg != '#') &&
+			     (*arg != '%') &&
 			     (*arg != '!') && (*arg != '*')))
 			{
 #ifdef NICK_LOG
@@ -1718,7 +1790,16 @@ static int scan_line(char *line, int isclient)
 			   	strncpy(user_nick, arg, sizeof(user_nick));
 			   	user_nick[sizeof(user_nick) - 1] = '\0';
 			}
+		   	if ((!motd_is_done) && 
+			    ((!strcasecmp(type,"376")) || 
+			     (!strcasecmp(type,"421"))))
+		     		motd_is_done = 1;
    		}
+	   	/* The server shouldn't start PINGing us until well
+		** after the end of the MOTD.. 
+		*/
+	   	if ((!motd_is_done) && (!strncasecmp(line,"PING",4)))
+	     		motd_is_done = 1;
      	}
    	else if (use_anonymity) 	/* Data is from client, anonymize */
      	{
@@ -1730,7 +1811,9 @@ static int scan_line(char *line, int isclient)
 	   	else if (!strncasecmp(line, "USER", 4))
 	     	{
 	   		int cip = ANON_USERID;
-	   		sprintf(line, "USER t%x anon anon :%s\n", cip, ANON_IRCNAME);
+	   		sprintf(line, 
+				"USER t%x anon anon :%s\n", 
+				cip, ANON_IRCNAME);
 		}
 	}
    	else if (!strncasecmp(line, "NOTICE", 6))	/* Fix mIRC junk */
@@ -1787,10 +1870,13 @@ static int scan_line(char *line, int isclient)
 			   		sprintf(out, ":%.32s %.256s", r, line);
 					quiz_delay_line(out, QUIZ_C);
 			   		quiz_greet();
+			   		retv = 1;
+			   		break;
 
 					/* Allow the USER command.
 			     		*/
 			 case 2:	retv = 1;
+			   		strcat(line,"#PiNG#\n");
 					break;
 
 				 	/* Check the users' responses.
@@ -1948,7 +2034,7 @@ static char *proxy_dcc(int destaddr, int destport, int incoming)
 	   	if (!okay) 
 	     	{
 		   	debug_msg(0, LOG_INFO,
-			       	"Dropped illegal DCC from: %128s (client %.32s@%.96s)",
+			       	"Dropped illegal DCC from: %.128s (client %.32s@%.96s)",
 			       		host_ip, user_name,
 			       		inet_ntoa(addr.sin_addr));
 		   	strcpy(retvalue,"junk junk");
@@ -1962,7 +2048,7 @@ static char *proxy_dcc(int destaddr, int destport, int incoming)
 	*/
 	len = 10;
 	while ((listen_sock = bind_to_port((incoming) ? visible_ip_o : visible_ip_i, 
-					   (rand() % 55555) + 1024 + getpid(), 
+					   ((rand() + getpid()) % PPOOL) + 1025, 
 					   1, DEBUG_TRIVIA)) < 0) 
      	{
 	   	if (--len < 1) {
@@ -2016,7 +2102,7 @@ static char *proxy_dcc(int destaddr, int destport, int incoming)
    	/* Give people five minutes to accept the call..
 	*/
 	signal(SIGALRM, alarm_signal);
-   	alarm(5*MIN);
+   	alarm(5*MINUTE);
    	strcpy(alarm_in,"proxy_dcc: accept");
 
    	/* No silly broadcasts shall mess up our DCC stuff!
@@ -2048,7 +2134,7 @@ static char *proxy_dcc(int destaddr, int destport, int incoming)
 	if ((to_addr.sin_addr.s_addr != clients_ip) &&
 	    (addr.sin_addr.s_addr != clients_ip))
      	{
-	   	debug_msg(0, LOG_WARNING, "Abuse: illegal DCC request %128s <-> %128s !",
+	   	debug_msg(0, LOG_WARNING, "Abuse: illegal DCC request %.128s <-> %.128s !",
 		       	       	inet_ntoa(to_addr.sin_addr),
 		       		inet_ntoa(addr.sin_addr));
      		exit(0);
@@ -2068,57 +2154,118 @@ static char *proxy_dcc(int destaddr, int destport, int incoming)
 }
 
 
-/* Change uid & gid!
+/* Read username from disk or shared memory, if possible.
 */
-static void change_uid(struct sockaddr_in *addr)
+static void get_user_name(struct sockaddr_in *addr)
 {
-#ifdef IP_TO_UID_PREFIX
-   	int		fd;
-   	int		tries;
-   	char 		ipfile[512];
-   	char		*cp;
-   	struct passwd	*pw;
-
-#ifdef CDIR
-	if (use_cdir) 
-      		sprintf(ipfile, "%s%s-%s", CDIR, CDIR_MAP, 
-			inet_ntoa(addr->sin_addr));
-   	else
-#endif     
-      		sprintf(ipfile, "%s%s", IP_TO_UID_PREFIX, 
-			inet_ntoa(addr->sin_addr));
-
-   	/*  Repeat until whe lose the ident race.. :-)
+   	int	tries = 0;
+#ifdef USE_UDB
+	struct udb_ip_user buf;
+	
+	/* People neither using the CDIR stuff nor running as root probably 
+	 * couldn't care less about ident problems.
 	 */
-	tries = 0;
-   	while ((fd = open(ipfile,O_RDONLY)) == -1)
-     	{
-	 	if (tries > 5) {
-	      		debug_msg(0, LOG_WARNING, 
-			       	"No %.256s file found, ident response will be wrong!", 
-			       	ipfile);
+	if (!use_cdir && getuid()) tries = 10;
+
+   	/* Be stubborn, to decrease the chance of race conditions.
+	 */
+	while (!udb_ip_get(&(addr->sin_addr), &buf))
+	{
+		if (tries > 3)
+		{
+			debug_msg(0, LOG_WARNING, 
+					  "Address not in UDB table, ident response will be wrong!");
+			return;
+		}
+		else
+		{
+			tries++;
+			sleep(1);
+		}
+	}
+	
+#ifdef TIRC_DEBUG
+	debug_msg(DEBUG_TRIVIA, LOG_DEBUG, "Got username: %.256s", buf.username);
+#endif
+
+	strncpy(user_name, buf.username, MAXLOGNAME);
+   	user_name[MAXLOGNAME-1] = '\0';
+	return;
+
+#endif /* USE_UDB */
+#ifdef IP_TO_UID_PREFIX /* not USE_UDB */
+   	int	 fd;
+   	char ipfile[PATH_MAX];
+   	char *cp;
+
+	/* People neither using the CDIR stuff nor running as root probably 
+	 * couldn't care less about ident problems.
+	 */
+	if (!use_cdir && getuid()) tries = 10;
+
+	/* FIXME: dangerous sprintfs */
+#ifdef CDIR
+	if (use_cdir)
+# ifdef HAVE_SNPRINTF
+	  snprintf(ipfile, sizeof(ipfile),
+# else
+	  sprintf(ipfile, 
+# endif
+			  "%s%s-%s", CDIR, CDIR_MAP, 
+			  inet_ntoa(addr->sin_addr));
+    else
+#endif
+# ifdef HAVE_SNPRINTF
+	  snprintf(ipfile, sizeof(ipfile),
+# else
+	  sprintf(ipfile, 
+# endif
+			  "%s%s", IP_TO_UID_PREFIX, 
+			  inet_ntoa(addr->sin_addr));
+
+   	/*  Repeat until we lose the ident race.. :-)
+	 */
+	while ((fd = open(ipfile,O_RDONLY)) == -1)
+	{
+		if (tries > 5) {
+			debug_msg(0, LOG_WARNING, 
+					  "No %.256s file found, ident response will be wrong!", 
+					  ipfile);
 		   	return;
-		} else {
-		   	sleep(1);
+		} 
+		else 
+		{
+			sleep(1);
 		}
 	   	tries++;
 	}
 
-   	user_name[0] = user_name[511] = '\0';
-   	read(fd, user_name, 511);
+   	user_name[0] = user_name[MAXLOGNAME-1] = '\0';
+   	read(fd, user_name, MAXLOGNAME-1);
    	close(fd);
 
    	if ((cp = strchr(user_name,'\n'))) *cp = '\0';
+	return;
+
+#endif /* IP_TO_UID_PREFIX */
+}
+	
+/* Change uid & gid!
+*/
+static void change_uid(void)
+{
+#if (defined IP_TO_UID_PREFIX || defined USE_UDB)
+   	struct passwd *pw;
 
    	/* We aren't running as root, so there's nothing more we can do..
 	*/
    	if (getuid()) return;
 
 	if ((pw = getpwnam(user_name)) == NULL)
-     	{
-      		debug_msg(0, LOG_WARNING, 
-		       	"Invalid user: %.32s, didn't changed UID/GID.", 
-		       	user_name);
+	{
+		debug_msg(0, LOG_WARNING, 
+				  "Invalid user: %.32s, didn't changed UID/GID.", 
+				  user_name);
 # ifdef PARANOID
 	   	if (use_paranoid) exit(0);
 # endif
@@ -2126,7 +2273,7 @@ static void change_uid(struct sockaddr_in *addr)
 	}
 
  	/* Change ids..
-      	*/
+	*/
    	setgid(pw->pw_gid);
    	setuid(pw->pw_uid);
 
@@ -2146,56 +2293,72 @@ static void change_uid(struct sockaddr_in *addr)
 #ifdef CDIR
 static void tell_identd(struct sockaddr_in *us, struct sockaddr_in *them)
 {
-   	int fd;
    	char fakeid[64], *u;
+# ifndef USE_UDB
+   	int fd;
    	*user_ident_file = '\0';
-
-   	/* If we don't have a valid username, just forget it.
-	*/
-   	if ((!*user_name) && (!use_anonymity)) return;
-
-      	sprintf(user_ident_file, "%s%s:%d-%s:%d", CDIR, CDIR_IDENT,
+	
+	sprintf(user_ident_file, 
+			"%s%s:%d-%s:%d", CDIR, CDIR_IDENT,
 			ntohs(us->sin_port),
 			inet_ntoa(them->sin_addr),
 			ntohs(them->sin_port));
-
+	
    	if ((fd = creat(user_ident_file, S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH)) == -1)
-     	{
-      		debug_msg(0, LOG_WARNING, 
-		       	"Unable to create %s, ident response will be wrong!", 
-		       	user_ident_file);
+	{
+		debug_msg(0, LOG_WARNING, 
+				  "Unable to create %s, ident response will be wrong!", 
+				  user_ident_file);
 	   	*user_ident_file = '\0';
 	   	return;
 	}
+# endif
 
-	if (use_anonymity)
-     	{
+	if ((use_anonymity) || (!*user_name))
+	{
 	   	int cip = ANON_USERID;
-
+		
 	   	sprintf(fakeid, "t%x", cip);
-      		debug_msg(0, LOG_INFO, "Anonymized connection as %s", fakeid);
-
-	       	u = fakeid;
+		debug_msg(0, LOG_INFO, "Anonymized connection as %s", fakeid);
+		
+		u = fakeid;
 	}
-     	else	u = user_name;
-   
+	else
+	  u = user_name;
+	
+# ifdef USE_UDB
+	memset(&conn, 0, sizeof(conn));
+	conn.from.sin_family = us->sin_family;
+	conn.from.sin_addr = us->sin_addr;
+	conn.from.sin_port = us->sin_port;
+	conn.to.sin_family = them->sin_family;
+	conn.to.sin_addr = them->sin_addr;	
+	conn.to.sin_port = them->sin_port;
+
+	udb_conn_record(&conn, u, "", UDB_OVERWRITE_LRU, UDB_NOW);
+	
+# else	
    	if (write(fd, u, strlen(u)) < 0)
-      		debug_msg(0, LOG_WARNING, 
+	  debug_msg(0, LOG_WARNING, 
 		       	"Unable to write to %s, ident response will be wrong!", 
 		       	user_ident_file);
 
    	close(fd);
+# endif
 }
 
 /* Cleanup ..
 */
 static void cleanup_identd(void)
 {
+# ifdef USE_UDB
+	udb_conn_delete(&conn);
+# else	
    	if (*user_ident_file)
      		unlink(user_ident_file);
-
+# endif
 }
-#endif
+#endif /* CDIR */
 
 
 /* Broadcast a message from sysadmin to proxy user.. It's the 
@@ -2288,6 +2451,7 @@ static void debug_msg(int lev, int pri, const char *format, ...) {
 	vsnprintf(msg, 511, format, ap);
 #else
 # if HAVE_VPRINTF
+#  warning Using vsprintf instead of vsnprintf!
 	vsprintf(msg, format, ap);
 # else
 #  error I need vsprintf or vsnprintf!
@@ -2310,19 +2474,25 @@ static void debug_msg(int lev, int pri, const char *format, ...) {
 
 /* Save a line that was destined for output for later..
 */
+int quiz_delayed = 0;
 static void quiz_delay_line(char *line, char **first, char ***bl)
 {
    	char *buf;
+	int linelen;
+	
+	quiz_delayed += linelen = strlen(line);
    
-	if (*bl == NULL) *bl = first;
+	if (quiz_delayed > QUIZ_MEMORY_HOG) exit(1);
 
-   	**bl = buf = malloc(sizeof(char *)+strlen(line)+1);
+	if (*bl == NULL) *bl = first;
+	
+   	**bl = buf = malloc(sizeof(char *)+linelen+1);
 	strcpy(QSTR(buf), line);
 
 	QPTR(buf) = NULL;
 	*bl = &QPTR(buf);
 
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 	debug_msg(DEBUG_TRIVIA, LOG_DEBUG, "Buffered: %.256s", line);
 #endif
 }
@@ -2340,14 +2510,14 @@ static void quiz_dump_lines(int sock, char **first, char ***bl)
 			      strlen(QSTR(*first))) < 0)
 			if (errno != EINTR)
 		{
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 			debug_msg(DEBUG_BASIC, LOG_WARNING, 
-			       "write(%s) failed in quiz_dump_lines: %128s",
+			       "write(%s) failed in quiz_dump_lines: %.128s",
 			       strerror(errno));
 #endif
 			return;
 		}
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 		debug_msg(DEBUG_TRIVIA, LOG_DEBUG, 
 			  "Flushed: %.256s", QSTR(*first));
 #endif
@@ -2404,7 +2574,7 @@ static void quiz_check_auth(char *pass)
 			}
 		}
 
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 		debug_msg(DEBUG_FEATURES, LOG_AUTH, 
 			  "Incorrect password given for %s", u);
 #endif
@@ -2435,7 +2605,7 @@ static void quiz_check_auth(char *pass)
 		}
 	   	else
 	     	{
-#ifdef DEBUG
+#ifdef TIRC_DEBUG
 			debug_msg(DEBUG_TRIVIA, LOG_DEBUG, 
 				  "Incorrect quiz response: %s", pass);
 #endif
@@ -2450,9 +2620,9 @@ static void quiz_check_auth(char *pass)
 static void quiz_msg(char *message)
 {
    	char out[512];
-
+	
    	sprintf(out,":%.32s!irc@proxy PRIVMSG %.32s :%.256s\n", 
-		QUIZ_NICK, user_nick, message);
+			QUIZ_NICK, user_nick, message);
 
    	quiz_delay_line(out, QUIZ_C);
 }
@@ -2461,56 +2631,60 @@ static void quiz_msg(char *message)
 */
 static void quiz_greet(void)
 {
-   	FILE		*fd;
-   	int		i, r;
-   	char 		question[512], buff[512], *p;
+   	FILE  *fd;
+   	int   i, r;
+   	char  question[512], buff[512], *p;
 
 	if (use_unix_passwd) 
-     	{
-     		quiz_msg(QUIZ_PASSWD_WAIT);
+	{
+		quiz_msg(QUIZ_PASSWD_WAIT);
 	   	return;
 	}
    	
    	/* Not using /etc/passwd, select a question from the quizfile */
-
+	
    	srand(time((time_t) NULL));
 
    	if ((fd = fopen(quizfile,"r")) == NULL)
-     	{
+	{
 	  	debug_msg(0, LOG_DEBUG, 
-			  "Error (%.128s) reading quiz-file: %.128s", 
-			  strerror(errno), quizfile);
+				  "Error (%.128s) reading quiz-file: %.128s", 
+				  strerror(errno), quizfile);
 	   	quiz_msg("Error opening quiz file.  Notify the proxy administrator.");
 	   	return;
 	}
-
+	
    	r = 0;
    	while (fgets(buff, 512, fd) != NULL) 
-     		if (*buff == '!') 
-		   	quiz_msg(buff+1);
-	   	else if (*buff != '#') 
-     		{
-			i = rand();
-		   	if ((i > r) || (!r)) 
-		     	{
-			   	r = i;
-			   	strcpy(question,buff);
-		     	}
+	  if (*buff == '!') 
+		quiz_msg(buff+1);
+	  else 
+	    if (*buff != '#') 
+	{
+		i = rand();
+		if ((i > r) || (!r)) 
+		{
+			r = i;
+			strcpy(question,buff);
 		}
-
+	}
+	
 	fclose(fd);
 
 	if ((p = strchr(question, ':')) != NULL)
-     	{
+	{
 	   	*p++ = '\0';
 	   	quiz_msg(question);
 	   	strcpy(quiz,p);
 	   	if ((p = strchr(quiz, '\n')) != NULL) *p = '\0';
 	} 
    	else
-     	{
-	  	debug_msg(0, LOG_DEBUG, "Invalid quiz: %.256s", question); 
-	   	quiz_msg("Invalid quiz in file.  Notify the proxy administrator.");
+	{
+		debug_msg(0, LOG_DEBUG, "Invalid quiz: %.256s", question); 
+		quiz_msg("Invalid quiz in file.  Notify the proxy administrator.");
  	}
 }
 #endif
+		   
+
+			   
